@@ -17,10 +17,11 @@
 
 #define CSTR2SYM(s) (ID2SYM(rb_intern((s))))
 
-extern void Init_redcarpet_rndr();
+void Init_redcarpet_rndr();
 
 struct redcarpet_renderopt {
 	struct html_renderopt html;
+	VALUE link_attributes;
 	VALUE self;
 	VALUE base_class;
 #ifdef HAVE_RUBY_ENCODING_H
